
/**
 * Implements an AVL tree.
 */

public class AvlTree {
  /** The tree root. */
        protected AvlNode root;
  /**
    * Construct the tree.
    */
   public AvlTree( ){
     root = null;
     }


        /**
         * Insert into the tree; duplicates are ignored.
         * @param x the item to insert.
         */
        public void insert( int x,  Object el )
        {
            root = insert( x, el, root );
        }

        /**
         * Remove from the tree. Nothing is done if x is not found.
         * @param x the item to remove.
         */
        public void remove( int x )
        {
            root = remove(x, root);
        }
        
        protected AvlNode remove(int x, AvlNode n) {
        	if(n==null) return null;
        	if(x==n.key) {
        		if(n.left==null && n.right==null) {
        			return null;
        		} else if(n.left == null) {
        			return n.right;
        		} else if(n.right == null) {
        			return n.left;
        		} else {
        			AvlNode pred = n.left;
        			while(pred.right != null)
        				pred = pred.right;
        			n.key = pred.key;
        			n.el = pred.el;
        			n.left = remove(n.key, n.left);
        			updateHeightAndSize(n);
							if(height(n.left)+1 < height(n.right)) {
								//n.right != null
								if(height(n.right.left) <= height(n.right.right)) {
									//Single rotation
									n = rotateWithRightChild(n);
								} else {
									//Double rotation
									n = doubleWithRightChild(n);
								}
							}
        			return n;	
        			
        		}
        	} else if(x<n.key) {
        		n.left = remove(x, n.left);
        		updateHeightAndSize(n);
        		if(height(n.left)+1 < height(n.right)) {
        			//n.right != null
        			if(height(n.right.left) <= height(n.right.right)) {
        				//Single rotation
        				n = rotateWithRightChild(n);
        			} else {
        				//Double rotation
        				n = doubleWithRightChild(n);
        			}
        		}
        		return n;
        	} else {
        		n.right = remove(x, n.right);
        		updateHeightAndSize(n);
        		if(height(n.right)+1 < height(n.left)) {
        			//n.left != null
        			if(height(n.left.right) <= height(n.left.left)) {
        				//Single rotation
        				n = rotateWithLeftChild(n);
        			} else {
        				//Double rotation
        				n = doubleWithLeftChild(n);
        			}
        		}
        		return n;
        	}
        }
        
        protected static void updateHeightAndSize(AvlNode n) {
        	n.height = 1 + max(height(n.left), height(n.right));
        	n.size = 1 + getSize(n.left) + getSize(n.right);
        }


        /**
         * Find an item in the tree.
         * @param x the item to search for.
         * @return the matching item or null if not found.
         */
        public Object find( int x )
        {
            AvlNode node = find( x, root );
            if(node!=null)
            	return node.el;
            else
            	return null;
        }

        /**
         * Make the tree logically empty.
         */
        public void makeEmpty( )
        {
            root = null;
        }

        /**
         * Test if the tree is logically empty.
         * @return true if empty, false otherwise.
         */
        public boolean isEmpty( )
        {
            return root == null;
        }

        /**
         * Print the tree contents in sorted order.
         */
        public void printTree( )
        {
            if( isEmpty( ) )
                System.out.println( "Empty tree" );
            else
                printTree( root );
        }

        /**
         * Internal method to get 	 field.
         * @param t the node.
         * @return the key field or null if t is null.
         */
        private int keyAt( AvlNode t )
        {
            return t == null ? null : t.key;
        }

        /**
         * Internal method to insert into a subtree.
         * @param x the item to insert.
         * @param t the node that roots the tree.
         * @return the new root.
         */
        private AvlNode insert( int x, Object el, AvlNode t )
        {
            if( t == null )
                t = new AvlNode( x, el, null, null );
            else if( x <  t.key )
            {
                t.left = insert( x, el, t.left );
                t.size++;
                if( height( t.left ) - height( t.right ) == 2 )
                    if( x < t.left.key )
                        t = rotateWithLeftChild( t );
                    else
                        t = doubleWithLeftChild( t );
            }
            else if( x >  t.key)
            {
                t.right = insert( x, el, t.right );
                t.size++;
                if( height( t.right ) - height( t.left ) == 2 )
                    if( x > t.right.key)
                        t = rotateWithRightChild( t );
                    else
                        t = doubleWithRightChild( t );
            }
            else
                ;  // Duplicate; do nothing
            t.height = max( height( t.left ), height( t.right ) ) + 1;
            return t;
        }


        /**
         * Internal method to find an item in a subtree.
         * @param x is item to search for.
         * @param t the node that roots the tree.
         * @return node containing the matched item.
         */
        private AvlNode find( int x, AvlNode t )
        {
            while( t != null )
                if( x < t.key  )
                    t = t.left;
                else if( x > t.key )
                    t = t.right;
                else
                    return t;    // Match

            return null;   // No match
        }

        /**
         * Internal method to print a subtree in sorted order.
         * @param t the node that roots the tree.
         */
        private void printTree( AvlNode t )
        {
            if( t != null )
            {
                printTree( t.left );
                System.out.println( t.key );
                printTree( t.right );
            }
        }


        /**
         * Return the height of node t, or -1, if null.
         * @param t node which height has to be compute.
         * @return
         */
        private static int height( AvlNode t )
        {
            return t == null ? -1 : t.height;
        }


        /**
         * Return maximum of lhs and rhs.
         * @param lhs
         * @param rhs
         * @return
         */
        private static int max( int lhs, int rhs )
        {
            return lhs > rhs ? lhs : rhs;
        }

        /**
         * Rotate binary tree node with left child.
         * For AVL trees, this is a single rotation for case 1.
         * Update heights, then return new root.
         */
        private static AvlNode rotateWithLeftChild( AvlNode k2 )
        {
            AvlNode k1 = k2.left;
            k2.left = k1.right;
            k1.right = k2;
            k2.height = max( height( k2.left ), height( k2.right ) ) + 1;
            k1.height = max( height( k1.left ), k2.height ) + 1;
            k1.size = k2.size;
            k2.size = getSize(k2.left) + getSize(k2.right) + 1;
            return k1;
        }

        /**
         * Rotate binary tree node with right child.
         * For AVL trees, this is a single rotation for case 4.
         * Update heights, then return new root.
         */
        private static AvlNode rotateWithRightChild( AvlNode k1 )
        {
            AvlNode k2 = k1.right;
            k1.right = k2.left;
            k2.left = k1;
            k1.height = max( height( k1.left ), height( k1.right ) ) + 1;
            k2.height = max( height( k2.right ), k1.height ) + 1;
            k2.size = k1.size;
            k1.size = getSize(k1.left) + getSize(k1.right) + 1;
            return k2;
        }
        
		  	protected static int getSize(AvlNode n) {
    			if(n==null)
    				return 0;
    			else
    				return n.size;
    		}


        /**
         * Double rotate binary tree node: first left child
         * with its right child; then node k3 with new left child.
         * For AVL trees, this is a double rotation for case 2.
         * Update heights, then return new root.
         */
        private static AvlNode doubleWithLeftChild( AvlNode k3 )
        {
            k3.left = rotateWithRightChild( k3.left );
            return rotateWithLeftChild( k3 );
        }

        /**
         * Double rotate binary tree node: first right child
         * with its left child; then node k1 with new right child.
         * For AVL trees, this is a double rotation for case 3.
         * Update heights, then return new root.
         */
        private static AvlNode doubleWithRightChild( AvlNode k1 )
        {
            k1.right = rotateWithLeftChild( k1.right );
            return rotateWithRightChild( k1 );
        }

				public void printDashes() {
					printDashes(root, "");
				}
				
				protected static void printDashes(AvlNode n, String dashes) {
					if(n==null) return;
					System.out.println(dashes+n.key+","+n.el);
					String newDashes = dashes+"-";
					printDashes(n.left, newDashes);
					printDashes(n.right, newDashes);
				}
				
				private class Quadruple {
					public int min;
					public int max;
					public int height;
					public int size;
				}
				
				public boolean isAvl() {
					return root==null || (isAvl(root) != null);
				}
				
				protected Quadruple isAvl(AvlNode n) {
					if(height(n.left)+1 < height(n.right) || height(n.right)+1 < height(n.left)) {
						return null;
					}
					Quadruple p = new Quadruple();
					p.min = p.max = n.key;
					p.height = 0;
					p.size = 1;
					if(n.left != null) {
						Quadruple pl = isAvl(n.left);
						if(pl == null || pl.max > n.key)
							return null;
						p.min = pl.min;
						p.height = pl.height + 1;
						p.size += pl.size;
					}
					if(n.right != null) {
						Quadruple pr = isAvl(n.right);
						if(pr == null || pr.min < n.key)
							return null;
						p.max = pr.max;
						if(p.height < pr.height + 1)
							p.height = pr.height + 1;
							p.size += pr.size;
					}
					if(p.height != n.height || p.size != n.size)
						return null;
					return p;
				}


}